import pytest
import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'pbix')))

from pbixray import PBIXRay

PBIX_FILE_PATH = r"C:\git\hub\pbixray\data\Sales & Returns Sample v201912.pbix"
# C:\git\hub\pbixray\data\Excalidraw.pbix

def test_initialization():
    """Test initialization of the library with the test PBIX file."""
    model = PBIXRay(PBIX_FILE_PATH)
    assert model is not None, "Failed to initialize PBIXRay with the test PBIX file."

def test_metadata_retrieval():
    """Test retrieval of metadata (tables, columns)."""
    model = PBIXRay(PBIX_FILE_PATH)
    
    tables = model.tables
    #test number of tables greater than 0
    assert tables.size, "No tables found in the PBIX file."
    
    assert "Age" in tables, "Expected table 'Age' not found in the PBIX file."
    
def test_data_retrieval():
    """Test data retrieval from a specific table."""
    model = PBIXRay(PBIX_FILE_PATH)
    
    table = model.get_table("Age")
    assert table.size, "Failed to retrieve the 'Age' table."



import xml.etree.ElementTree as ET
from .backup_log import BackupLog
from .backup_log_header import BackupLogHeader
from .virtual_directory import VirtualDirectory
from .data_model import DataModel

class AbfParser:
    def __init__(self, data_model:DataModel):
        self.data_model = data_model
        self.__buffer = data_model.decompressed_data
        self.__backup_log_header = None
        self.__virtual_directory = None
        self.__backup_log = None
        self.file_log = None  # public attribute

        # Trigger the parsing process upon initialization
        self.__parse_all()

    def __parse_all(self):
        self.__parse_backup_log_header()
        self.__parse_virtual_directory()
        self.__parse_backup_log()
        self.__match_logs_and_get_attributes()

    def __parse_backup_log_header(self):
        offset = 72 # STREAM_STORAGE_SIGNATURE_)!@#$%^&*(
        page = 0x1000 # header is always one page (4096 bytes) in size
        self.__backup_log_header = BackupLogHeader(self.__buffer[offset:page])

    def __parse_virtual_directory(self):
        offset = int(self.__backup_log_header.m_cbOffsetHeader)
        size = int(self.__backup_log_header.DataSize)
        self.__virtual_directory = VirtualDirectory(self.__buffer[offset: offset+size])

    def __parse_backup_log(self):
        log = self.__virtual_directory.BackupFiles[-1]
        offset = int(log.m_cbOffsetHeader)
        size = int(log.Size)
        self.data_model.error_code = self.__backup_log_header.ErrorCode
        self.__backup_log = BackupLog(self.__buffer[offset:offset+size], self.__backup_log_header.ErrorCode)

    def __match_logs_and_get_attributes(self):
        persist_root = self.__backup_log.FileGroups[1].PersistLocationPath + '\\'
        virtual_directory_files_by_path = {file.Path: file for file in self.__virtual_directory.BackupFiles}
        matched_data = []

        for file_group in self.__backup_log.FileGroups:
            for backup_file in file_group.FileList:
                if backup_file.StoragePath in virtual_directory_files_by_path:
                    matched_file = virtual_directory_files_by_path[backup_file.StoragePath]
                    if backup_file.Path.startswith(persist_root):
                        path_without_persist_root = backup_file.Path.replace(persist_root, '', 1)
                    else:
                        path_without_persist_root = backup_file.Path
                    matched_data.append({
                        'Path': path_without_persist_root,
                        'FileName': path_without_persist_root.split('\\')[-1],
                        'StoragePath': backup_file.StoragePath,
                        'Size': matched_file.Size,
                        'm_cbOffsetHeader': matched_file.m_cbOffsetHeader
                    })

        self.data_model.file_log = matched_data



class HuffmanTree:
    def __init__(self, c=0):
        self.c = c
        self.left = None
        self.right = None

# Helper function to convert ISO-8859-1 code to UTF-8 string
def iso88591_to_utf8(code):
    if code >= 0x80:
        return bytes([0xC2 + (code > 0xBF), (code & 0x3F) + 0x80]).decode('utf-8')
    else:
        return chr(code)

# Function to generate the full 256-byte Huffman array from the compact 128-byte encode_array
def decompress_encode_array(compressed):
    full_array = [0] * 256
    for i, byte in enumerate(compressed):
        full_array[2 * i] = byte & 0x0F         # Lower nibble
        full_array[2 * i + 1] = (byte >> 4) & 0x0F  # Upper nibble
    return full_array

# Function to generate Huffman codes based on codeword lengths
def generate_codes(lengths):
    codes = {}
    sorted_lengths = [(lengths[i], i) for i in range(256) if lengths[i] != 0]
    sorted_lengths.sort()

    code = 0
    last_length = 0
    for length, character in sorted_lengths:
        if last_length != length:
            code <<= (length - last_length)
            last_length = length
        codes[character] = bin(code)[2:].zfill(length)
        code += 1
    return codes

# Function to build Huffman tree based on generated codes
def build_huffman_tree(encode_array):
    codes = generate_codes(encode_array)
    root = HuffmanTree()
    for character, code in codes.items():
        node = root
        for bit in code:
            if bit == '0':
                if not node.left:
                    node.left = HuffmanTree()
                node = node.left
            else:
                if not node.right:
                    node.right = HuffmanTree()
                node = node.right
        node.c = character
    return root

# Function to decode a bitstream from start to end bit positions using the Huffman tree
def decode_substring(bitstream, tree, start_bit, end_bit):
    result = ""
    node = tree
    total_bits = end_bit - start_bit
    for i in range(total_bits):
        bit_pos = start_bit + i
        byte_pos = bit_pos // 8
        bit_offset = bit_pos % 8
        byte_pos = (byte_pos & ~0x01) + (1 - (byte_pos & 0x01))

        if not node.left and not node.right:
            result += iso88591_to_utf8(node.c)
            node = tree

        if bitstream[byte_pos] & (1 << (7 - bit_offset)):
            node = node.right
        else:
            node = node.left

    if not node.left and not node.right:
        result += iso88591_to_utf8(node.c)

    return result



import xml.etree.ElementTree as ET

class VirtualDirectoryBackupFile:
    def __init__(self, element):
        self.Path = element.findtext("Path")
        self.Size = int(element.findtext("Size"))
        self.m_cbOffsetHeader = int(element.findtext("m_cbOffsetHeader"))
        self.Delete = element.findtext("Delete") == "true"
        self.CreatedTimestamp = int(element.findtext("CreatedTimestamp"))
        self.Access = int(element.findtext("Access"))
        self.LastWriteTime = int(element.findtext("LastWriteTime"))

class VirtualDirectory:
    def __init__(self, xml_string):
        root = ET.fromstring(xml_string)
        self.BackupFiles = [VirtualDirectoryBackupFile(e) for e in root.findall("BackupFile")]

    @classmethod
    def from_xml_string(cls, xml_string):
        return cls(xml_string)




class MetadataQuery:
    def __init__(self, sqlite_handler):
        self.handler = sqlite_handler
        
        # Populate dataframes upon instantiation
        self.schema_df = self.__populate_schema()
        self.m_df = self.__populate_m()
        self.m_parameters_df = self.__populate_m_parameters()
        self.dax_tables_df = self.__populate_dax_tables()
        self.dax_measures_df = self.__populate_dax_measures()
        self.dax_columns_df = self.__populate_dax_columns()
        self.metadata_df = self.__populate_metadata()
        self.relationships_df = self.__populate_relationships()
        self.handler.close_connection()

    def __populate_schema(self):
        sql = """ 
        SELECT 
            t.Name AS TableName,
            c.ExplicitName AS ColumnName,
            sfd.FileName AS Dictionary, 
            sfh.FileName AS HIDX, 
            sfi.FileName AS IDF,
            cs.Statistics_DistinctStates as Cardinality,
            c.ExplicitDataType AS DataType,
            --ds.DataType,
            ds.BaseId,
            ds.Magnitude,
            ds.IsNullable,
            c.ModifiedTime,
            c.StructureModifiedTime
        FROM Column c 
        JOIN [Table] t ON c.TableId = t.ID
        JOIN ColumnStorage cs ON c.ColumnStorageID = cs.ID
        --HIDX
        JOIN AttributeHierarchy ah ON ah.ColumnID = c.ID
        JOIN AttributeHierarchyStorage ahs ON ah.AttributeHierarchyStorageID = ahs.ID
        LEFT JOIN StorageFile sfh ON sfh.ID = ahs.StorageFileID
        --Dictionary
        LEFT JOIN DictionaryStorage ds ON ds.ID = cs.DictionaryStorageID
        LEFT JOIN StorageFile sfd ON sfd.ID = ds.StorageFileID
        --IDF
        JOIN ColumnPartitionStorage cps ON cps.ColumnStorageID = cs.ID
        JOIN StorageFile sfi ON sfi.ID = cps.StorageFileID
        WHERE c.Type IN (1,2)
        ORDER BY t.Name, cs.StoragePosition
        """
        return self.handler.execute_query(sql)

    def __populate_m(self):
        sql = """ 
        SELECT 
            t.Name AS 'TableName', 
            p.QueryDefinition AS 'Expression'
        FROM partition p 
        JOIN [Table] t ON t.ID = p.TableID 
        WHERE p.Type = 4;
        """
        return self.handler.execute_query(sql)
    
    def __populate_m_parameters(self):
        sql = """ 
        SELECT 
            Name as ParameterName, 
            Description, 
            Expression, 
            ModifiedTime 
        FROM Expression;
        """
        return self.handler.execute_query(sql)  

    def __populate_dax_tables(self):
        sql = """ 
        SELECT 
            t.Name AS 'TableName', 
            p.QueryDefinition AS 'Expression'
        FROM partition p 
        JOIN [Table] t ON t.ID = p.TableID 
        WHERE p.Type = 2;
        """
        return self.handler.execute_query(sql)

    def __populate_dax_measures(self):
        sql = """ 
        SELECT 
            t.Name AS TableName,
            m.Name,
            m.Expression,
            m.DisplayFolder,
            m.Description
        FROM Measure m 
        JOIN [Table] t ON m.TableID = t.ID;
        """
        return self.handler.execute_query(sql)
    
    def __populate_dax_columns(self):
        sql = """ 
        SELECT 
            t.Name AS TableName,
            c.ExplicitName AS ColumnName,
            c.Expression
        FROM Column c 
        JOIN [Table] t ON c.TableID = t.ID
        WHERE c.Type = 2;
        """
        return self.handler.execute_query(sql)

    def __populate_metadata(self):
        sql = """
        SELECT Name,Value 
        FROM Annotation 
        WHERE ObjectType = 1
        """
        return self.handler.execute_query(sql)
    
    def __populate_relationships(self):
        sql = """
        SELECT 
            ft.Name AS FromTableName,
            fc.ExplicitName AS FromColumnName,
            tt.Name AS ToTableName,
            tc.ExplicitName AS ToColumnName,
            rel.IsActive,
            CASE 
                WHEN rel.FromCardinality = 2 THEN 'M'
                ELSE '1'
            END || ':' || 
            CASE 
                WHEN rel.ToCardinality = 2 THEN 'M'
                ELSE '1'
            END AS Cardinality,
            CASE 
                WHEN rel.CrossFilteringBehavior = 1 THEN 'Single'
                WHEN rel.CrossFilteringBehavior = 2 THEN 'Both'
                ELSE CAST(rel.CrossFilteringBehavior AS TEXT)
            END AS CrossFilteringBehavior,
            rid.RecordCount as FromKeyCount,
            rid2.RecordCount AS ToKeyCount,
            rel.RelyOnReferentialIntegrity
        FROM Relationship rel
            LEFT JOIN [Table] ft ON rel.FromTableID = ft.id
            LEFT JOIN [Column] fc ON rel.FromColumnID = fc.id
            LEFT JOIN [Table] tt ON rel.ToTableID = tt.id AND tt.systemflags = 0
            LEFT JOIN [Column] tc ON rel.ToColumnID = tc.id
            LEFT JOIN RelationshipStorage rs ON rs.id = rel.RelationshipStorageID
            LEFT JOIN RelationshipIndexStorage rid ON rs.RelationshipIndexStorageID = rid.id
            LEFT JOIN RelationshipStorage rs2 ON rs2.id = rel.RelationshipStorage2ID
            LEFT JOIN RelationshipIndexStorage rid2 ON rs2.RelationshipIndexStorageID = rid2.id
        """
        return self.handler.execute_query(sql)


# ---------- IMPORTS ----------
from .column_data.idf import ColumnDataIdf
from .column_data.idfmeta import IdfmetaParser
from .column_data.hidx import ColumnDataHidx
from .column_data.dictionary import ColumnDataDictionary
from .abf.backup_log import BackupLog
from .abf.virtual_directory import VirtualDirectory
from kaitaistruct import KaitaiStream
from .utils import AMO_PANDAS_TYPE_MAPPING, get_data_slice
import io
import pandas as pd
from decimal import Decimal
from .abf.data_model import DataModel

from .huffman import decompress_encode_array,build_huffman_tree, decode_substring
from collections import defaultdict

# ---------- VertiPaq CLASS ----------

class VertiPaqDecoder:
    def __init__(self, metadata, data_model:DataModel):
        self._meta = metadata
        self._data_model = data_model

    def _read_bitpacked(self,sub_segment, bit_width, min_data_id):
        """Reads bitpacked values from a sub_segment."""
        mask = (1 << bit_width) - 1
        res = []
        for u8le in sub_segment:
            for _ in range(64 // bit_width):
                res.append((min_data_id + (u8le & mask)))
                u8le >>= bit_width

        return res

    def _extract_strings(self,buffer):
        """Extract zero-terminated strings from buffer."""
        strings = buffer.split('\0')
        return strings[:-1]  # remove the last empty string

    def _read_rle_bit_packed_hybrid(self,buffer, entries, min_data_id, bit_width ):
        """Reads RLE bit packed hybrid values from a buffer."""
        with io.BytesIO(buffer) as f:
            # Parse the binary data
            column_data = ColumnDataIdf(KaitaiStream(f))
            
            bitpacked_values = []
            vector = []
            bit_packed_entries = None
            bit_packed_offset = 0
            
            if entries > 0:
                # Get bit width
                size = column_data.segments[0].sub_segment_size
                # case if it's a column with empty strings
                if column_data.segments[0].sub_segment[-1].bit_length() == 0 and size == 1:
                    bitpacked_values = [min_data_id] * entries
                else:
                    # read the bitpacked values from the sub_segment
                    bitpacked_values = self._read_bitpacked(column_data.segments[0].sub_segment,bit_width, min_data_id)

            # consider only the first primary segment + sub segment combination
            # for segment in column_data.segments: 
            for entry in column_data.segments[0].primary_segment:
                if entry.data_value+bit_packed_offset== 0xFFFFFFFF: # bit pack marker
                    bit_packed_entries = entry.repeat_value
                    bitpacked_values_slice = bitpacked_values[bit_packed_offset:bit_packed_offset+bit_packed_entries]
                    bit_packed_offset += bit_packed_entries
                    vector+=bitpacked_values_slice
                else:
                    rle = [entry.data_value] * entry.repeat_value
                    vector+=rle

            return vector

    def _read_idfmeta(self,buffer):
        """Reads idfmeta from a buffer."""
        # Use io.BytesIO to wrap the bytearray
        with io.BytesIO(buffer) as f:
            metadata = IdfmetaParser.from_io(f)
            
            # Extract the necessary data from the Kaitai Struct
            row_data = {
                'min_data_id': metadata.blocks.cp.cs.ss.min_data_id,
                'count_bit_packed': metadata.blocks.cp.cs.cs.count_bit_packed,
                'bit_width': metadata.bit_width,
            }
            
            return row_data

    def _read_hash_table(self,buffer):
        """Reads a hash table from a buffer."""
        with io.BytesIO(buffer) as f:
            # Parse the .hidx file using the Kaitai Struct
            parsed_hidx = ColumnDataHidx.from_io(f)

            # Create a hash table to store the results
            result_hash_table = {}

            # Iterate over each hash_bin entry
            for hash_bin in parsed_hidx.hash_bin_entries:
                # Iterate over each m_rg_local_entry inside hash_bin
                for local_entry in hash_bin.m_rg_local_entries:
                    # If the m_hash is non-zero, add it to the result hash table
                    if local_entry.m_hash != 0:
                        result_hash_table[local_entry.m_hash] =  local_entry.m_key
            
            # Iterate over the overflow_hash_entries
            for overflow_entry in parsed_hidx.overflow_hash_entries:
                # If the m_hash is non-zero, add it to the result hash table
                if overflow_entry.m_hash != 0:
                    result_hash_table[overflow_entry.m_hash] = overflow_entry.m_key 

            return result_hash_table

 
    def _read_dictionary(self, buffer, min_data_id):
        """Reads a dictionary from a buffer."""
        with io.BytesIO(buffer) as f:
            dictionary = ColumnDataDictionary.from_io(f)

        if dictionary.dictionary_type == ColumnDataDictionary.DictionaryTypes.xm_type_string:
            hashtable = {}
            index = min_data_id

            pages = dictionary.data.dictionary_pages
            record_handles = dictionary.data.dictionary_record_handles_vector_info.vector_of_record_handle_structures
            record_handles_map = defaultdict(list)

            for handle in record_handles:
                record_handles_map[handle.page_id].append(handle.bit_or_byte_offset)

            for page_id, page in enumerate(pages):
                if page.page_compressed:
                    compressed_store = page.string_store
                    encode_array = compressed_store.encode_array
                    store_total_bits = compressed_store.store_total_bits
                    compressed_string_buffer = compressed_store.compressed_string_buffer
                    ui_decode_bits = compressed_store.ui_decode_bits

                    full_encode_array = decompress_encode_array(encode_array)
                    huffman_tree = build_huffman_tree(full_encode_array)

                    if page_id in record_handles_map:
                        offsets = record_handles_map[page_id]
                        for i in range(len(offsets)):
                            start_bit = offsets[i]
                            end_bit = offsets[i + 1] if i + 1 < len(offsets) else store_total_bits
                            decompressed = decode_substring(compressed_string_buffer, huffman_tree, start_bit, end_bit)
                            hashtable[index] = decompressed
                            index += 1
                    del huffman_tree
                else:
                    uncompressed_store = page.string_store
                    uncompressed = uncompressed_store.uncompressed_character_buffer
                    strings = self._extract_strings(uncompressed)
                    for i, token in enumerate(strings):
                        hashtable[index] = token
                        index += 1

            return hashtable
        elif dictionary.dictionary_type in [ColumnDataDictionary.DictionaryTypes.xm_type_long, ColumnDataDictionary.DictionaryTypes.xm_type_real]:
            vector_values = dictionary.data.vector_of_vectors_info.values
            return {i: val for i, val in enumerate(vector_values, start=min_data_id)}

        return None    
        
    def _get_column_data(self, column_metadata, meta):
        """Extracts column data based on the given column metadata and meta information."""
        if pd.notnull(column_metadata["Dictionary"]):
            dictionary_buffer = get_data_slice(self._data_model,column_metadata["Dictionary"])
            null_adjustment = 1 if column_metadata["IsNullable"] else 0
            # Read and construct the dictionary with appropriate minimum data ID
            min_data_id_adj = meta['min_data_id'] - null_adjustment
            dictionary = self._read_dictionary(dictionary_buffer, min_data_id=meta['min_data_id'])
            data_slice = get_data_slice(self._data_model,column_metadata["IDF"])
            return pd.Series(self._read_rle_bit_packed_hybrid(data_slice, meta['count_bit_packed'], min_data_id_adj , meta['bit_width'])).map(dictionary)
        elif pd.notnull(column_metadata["HIDX"]):
            data_slice = get_data_slice(self._data_model,column_metadata["IDF"])
            return pd.Series(self._read_rle_bit_packed_hybrid(data_slice, meta['count_bit_packed'], meta['min_data_id'], meta['bit_width'])).add(column_metadata["BaseId"]) / column_metadata["Magnitude"]
        else:
            raise ValueError(f"Neither dictionary nor hidx found for column {column_metadata['ColumnName']} in table.")
        
    def _handle_special_cases(self, column_data, data_type):
        if data_type == 9:
            # Convert to datetime
            return pd.to_datetime(column_data, unit='d', origin='1899-12-30')
        elif data_type == 10:
            # Handle decimal.Decimal type
            return column_data.apply(lambda x: Decimal(x)/10000 if pd.notnull(x) else None)
        return column_data
        
    def get_table(self, table_name):
        """Generates a DataFrame representation of the specified table."""
        table_metadata_df = self._meta.schema_df[self._meta.schema_df['TableName'] == table_name]
        dataframe_data = {}

        for _, column_metadata in table_metadata_df.iterrows():
            idfmeta_buffer = get_data_slice(self._data_model,column_metadata["IDF"] + 'meta')
            meta = self._read_idfmeta(idfmeta_buffer)
            
            column_data = self._get_column_data(column_metadata, meta)
            # Handle special cases for certain data types
            column_data = self._handle_special_cases(column_data, column_metadata["DataType"])
            
            pandas_dtype = AMO_PANDAS_TYPE_MAPPING.get(column_metadata["DataType"], "object")  # default to object if no mapping is found
            
            # If it's a decimal type, keep it as object since pandas doesn't support Decimal natively
            if pandas_dtype == 'decimal.Decimal':
                pandas_dtype = 'object'
            dataframe_data[column_metadata["ColumnName"]] = column_data.astype(pandas_dtype)

        return pd.DataFrame(dataframe_data)


# ---------- IMPORTS ----------

from .pbix_unpacker import PbixUnpacker
from .vertipaq_decoder import VertiPaqDecoder
from .meta.metadata_handler import MetadataHandler
from .utils import WINDOWS_EPOCH_START
import datetime

# ---------- MAIN CLASS ----------

class PBIXRay:
    def __init__(self, file_path):
        unpacker = PbixUnpacker(file_path)
        
        self._metadata_handler = MetadataHandler(unpacker.data_model)
        self._vertipaq_decoder = VertiPaqDecoder(self._metadata_handler.metadata, unpacker.data_model)
        
    def get_table(self, table_name):
        """Generates a DataFrame representation of the specified table."""
        return self._vertipaq_decoder.get_table(table_name)

    # ---------- PROPERTIES ----------

    @property   
    def tables(self):
        return self._metadata_handler.tables
    
    @property
    def statistics(self):
        return self._metadata_handler.stats
    
    @property
    def power_query(self):
        return self._metadata_handler.metadata.m_df
    
    @property
    def m_parameters(self):
        df = self._metadata_handler.metadata.m_parameters_df
        return df if df.empty else df.assign(
            ModifiedTime=lambda df: df['ModifiedTime'].apply(
                lambda x: WINDOWS_EPOCH_START + datetime.timedelta(seconds=x / 1e7)
            )
        )

    @property
    def dax_tables(self):
        return self._metadata_handler.metadata.dax_tables_df
    
    @property
    def dax_measures(self):
        return self._metadata_handler.metadata.dax_measures_df
    
    @property
    def dax_columns(self):
        return self._metadata_handler.metadata.dax_columns_df
    
    @property
    def metadata(self):
        return self._metadata_handler.metadata.metadata_df
    
    @property
    def size(self):
        return self._metadata_handler.size
    
    @property
    def schema(self):
        return  self._metadata_handler.schema
    
    @property
    def relationships(self):
        return self._metadata_handler.metadata.relationships_df



# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO


if getattr(kaitaistruct, 'API_VERSION', (0, 9)) < (0, 9):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class ColumnDataHidx(KaitaiStruct):
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._read()

    def _read(self):
        self.hash_algorithm = self._io.read_s4le()
        self.hash_entry_size = self._io.read_u4le()
        self.hash_bin_size = self._io.read_u4le()
        self.local_entry_count = self._io.read_u4le()
        self.c_bins = self._io.read_u8le()
        self.number_of_records = self._io.read_s8le()
        self.current_mask = self._io.read_s8le()
        self.hash_stats = self._io.read_u1()
        if self.hash_stats != 0:
            self.hash_statistics = ColumnDataHidx.HashStatisticsType(self._io, self, self._root)

        self._raw_hash_bin_entries = []
        self.hash_bin_entries = []
        for i in range(self.c_bins):
            self._raw_hash_bin_entries.append(self._io.read_bytes(self.hash_bin_size))
            _io__raw_hash_bin_entries = KaitaiStream(BytesIO(self._raw_hash_bin_entries[i]))
            self.hash_bin_entries.append(ColumnDataHidx.HashBin(_io__raw_hash_bin_entries, self, self._root))

        self.overflow_hash_entries_count = self._io.read_u8le()
        self.overflow_hash_entries = []
        for i in range(self.overflow_hash_entries_count):
            self.overflow_hash_entries.append(ColumnDataHidx.HashEntry(self._io, self, self._root))


    class HashStatisticsType(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.number_of_elements = self._io.read_u8le()
            self.number_of_bins = self._io.read_u8le()
            self.number_of_used_bins = self._io.read_u8le()
            self.fast_access_elements = self._io.read_u8le()
            self.locals_size_per_bin = self._io.read_u8le()
            self.maximum_chain = self._io.read_u8le()
            self.element_count = self._io.read_u8le()
            self.element_size = self._io.read_u4le()
            self.histogram_vector = []
            for i in range(self.element_count):
                _on = self.element_size
                if _on == 4:
                    self.histogram_vector.append(self._io.read_u4le())
                elif _on == 8:
                    self.histogram_vector.append(self._io.read_u8le())



    class HashBin(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.m_rg_chain = self._io.read_u8le()
            self.m_count = self._io.read_u4le()
            self.m_rg_local_entries = []
            for i in range(self._root.local_entry_count):
                self.m_rg_local_entries.append(ColumnDataHidx.HashEntry(self._io, self, self._root))

            self.padding = self._io.read_u4le()


    class HashEntry(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.m_hash = self._io.read_u4le()
            self.m_key = self._io.read_u4le()






import apsw
import pandas as pd
import warnings

class SQLiteHandler:
    def __init__(self, sqlite_buffer):
        self.sqlite_buffer = sqlite_buffer
        self.conn = self._setup_connection()
        
    def _setup_connection(self):
        """Set up an in-memory SQLite connection and deserialize."""
        conn = apsw.Connection(":memory:")
        conn.deserialize("main", self.sqlite_buffer)
        return conn

    def execute_query(self, sql):
        """Execute a SQL query and return the result as a pandas DataFrame."""
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            try:
                return pd.read_sql_query(sql, self.conn)
            except apsw.ExecutionCompleteError:  
                return pd.DataFrame()  # Return an empty DataFrame for ExecutionCompleteError
            except apsw.SQLError as e:  # Handle other SQLErrors
                print(f"SQL error: {e}")
                return pd.DataFrame()  # Optional: Return an empty DataFrame or you can raise the error again

    def close_connection(self):
        """Close the SQLite connection."""
        self.conn.close()



# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
from enum import Enum


if getattr(kaitaistruct, 'API_VERSION', (0, 9)) < (0, 9):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class ColumnDataDictionary(KaitaiStruct):

    class DictionaryTypes(Enum):
        xm_type_invalid = -1
        xm_type_long = 0
        xm_type_real = 1
        xm_type_string = 2
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._read()

    def _read(self):
        self.dictionary_type = KaitaiStream.resolve_enum(ColumnDataDictionary.DictionaryTypes, self._io.read_s4le())
        self.hash_information = ColumnDataDictionary.HashInfo(self._io, self, self._root)
        _on = self.dictionary_type
        if _on == ColumnDataDictionary.DictionaryTypes.xm_type_string:
            self.data = ColumnDataDictionary.StringData(self._io, self, self._root)
        elif _on == ColumnDataDictionary.DictionaryTypes.xm_type_long:
            self.data = ColumnDataDictionary.NumberData(self._io, self, self._root)
        elif _on == ColumnDataDictionary.DictionaryTypes.xm_type_real:
            self.data = ColumnDataDictionary.NumberData(self._io, self, self._root)

    class StringRecordHandle(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.bit_or_byte_offset = self._io.read_u4le()
            self.page_id = self._io.read_u4le()


    class StringData(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.page_layout_information = ColumnDataDictionary.PageLayout(self._io, self, self._root)
            self.dictionary_pages = []
            for i in range(self.page_layout_information.store_page_count):
                self.dictionary_pages.append(ColumnDataDictionary.DictionaryPage(self._io, self, self._root))

            self.dictionary_record_handles_vector_info = ColumnDataDictionary.DictionaryRecordHandlesVector(self._io, self, self._root)


    class HashInfo(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.hash_elements = []
            for i in range(6):
                self.hash_elements.append(self._io.read_s4le())



    class VectorOfVectors(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.element_count = self._io.read_u8le()
            self.element_size = self._io.read_u4le()
            self.values = []
            for i in range(self.element_count):
                _on = self.data_type_id
                if _on == u"int32":
                    self.values.append(self._io.read_s4le())
                elif _on == u"int64":
                    self.values.append(self._io.read_s8le())
                elif _on == u"float64":
                    self.values.append(self._io.read_f8le())


        @property
        def is_int32(self):
            if hasattr(self, '_m_is_int32'):
                return self._m_is_int32

            self._m_is_int32 = self.element_size == 4
            return getattr(self, '_m_is_int32', None)

        @property
        def is_int64(self):
            if hasattr(self, '_m_is_int64'):
                return self._m_is_int64

            self._m_is_int64 =  ((self.element_size == 8) and (self._root.dictionary_type == ColumnDataDictionary.DictionaryTypes.xm_type_long)) 
            return getattr(self, '_m_is_int64', None)

        @property
        def is_float64(self):
            if hasattr(self, '_m_is_float64'):
                return self._m_is_float64

            self._m_is_float64 =  ((self.element_size == 8) and (self._root.dictionary_type == ColumnDataDictionary.DictionaryTypes.xm_type_real)) 
            return getattr(self, '_m_is_float64', None)

        @property
        def data_type_id(self):
            if hasattr(self, '_m_data_type_id'):
                return self._m_data_type_id

            self._m_data_type_id = (u"int32" if self.is_int32 else (u"int64" if self.is_int64 else u"float64"))
            return getattr(self, '_m_data_type_id', None)


    class CompressedStrings(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.store_total_bits = self._io.read_u4le()
            self.character_set_type_identifier = self._io.read_u4le()
            self.allocation_size = self._io.read_u8le()
            self.character_set_used = self._io.read_u1()
            self.ui_decode_bits = self._io.read_u4le()
            self.encode_array = []
            for i in range(128):
                self.encode_array.append(self._io.read_u1())

            self.ui64_buffer_size = self._io.read_u8le()
            self.compressed_string_buffer = self._io.read_bytes(self.allocation_size)


    class PageLayout(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.store_string_count = self._io.read_s8le()
            self.f_store_compressed = self._io.read_s1()
            self.store_longest_string = self._io.read_s8le()
            self.store_page_count = self._io.read_s8le()


    class DictionaryPage(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.page_mask = self._io.read_u8le()
            self.page_contains_nulls = self._io.read_u1()
            self.page_start_index = self._io.read_u8le()
            self.page_string_count = self._io.read_u8le()
            self.page_compressed = self._io.read_u1()
            self.string_store_begin_mark = self._io.read_bytes(4)
            if not self.string_store_begin_mark == b"\xDD\xCC\xBB\xAA":
                raise kaitaistruct.ValidationNotEqualError(b"\xDD\xCC\xBB\xAA", self.string_store_begin_mark, self._io, u"/types/dictionary_page/seq/5")
            _on = self.page_compressed
            if _on == 0:
                self.string_store = ColumnDataDictionary.UncompressedStrings(self._io, self, self._root)
            elif _on == 1:
                self.string_store = ColumnDataDictionary.CompressedStrings(self._io, self, self._root)
            self.string_store_end_mark = self._io.read_bytes(4)
            if not self.string_store_end_mark == b"\xCD\xAB\xCD\xAB":
                raise kaitaistruct.ValidationNotEqualError(b"\xCD\xAB\xCD\xAB", self.string_store_end_mark, self._io, u"/types/dictionary_page/seq/7")


    class OtherRecordHandle(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.bit_or_byte_offset = self._io.read_u4le()


    class UncompressedStrings(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.remaining_store_available = self._io.read_u8le()
            self.buffer_used_characters = self._io.read_u8le()
            self.allocation_size = self._io.read_u8le()
            self.uncompressed_character_buffer = (self._io.read_bytes(self.allocation_size)).decode(u"UTF-16LE")


    class NumberData(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.vector_of_vectors_info = ColumnDataDictionary.VectorOfVectors(self._io, self, self._root)


    class DictionaryRecordHandlesVector(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.element_count = self._io.read_u8le()
            self.element_size = self._io.read_bytes(4)
            if not self.element_size == b"\x08\x00\x00\x00":
                raise kaitaistruct.ValidationNotEqualError(b"\x08\x00\x00\x00", self.element_size, self._io, u"/types/dictionary_record_handles_vector/seq/1")
            self.vector_of_record_handle_structures = []
            for i in range(self.element_count):
                self.vector_of_record_handle_structures.append(ColumnDataDictionary.StringRecordHandle(self._io, self, self._root))







from .metadata_query import MetadataQuery
from .sqlite_handler import SQLiteHandler
from ..utils import AMO_PANDAS_TYPE_MAPPING, WINDOWS_EPOCH_START, get_data_slice
import pandas as pd
from ..abf.data_model import DataModel
import datetime

# ---------- METADATA HANDLER ----------

class MetadataHandler:
    def __init__(self, data_model:DataModel):
        self._data_model=data_model
        self._load_metadata()
        self._compute_statistics()
        
    def _load_metadata(self):
        """Loads metadata for the given PBIX file."""
        sqliteBuffer = get_data_slice(self._data_model,'metadata.sqlitedb')
        sqliteHandler = SQLiteHandler(sqliteBuffer)
        self._meta = MetadataQuery(sqliteHandler)
    
    def _compute_statistics(self):
        """Computes statistics from the metadata schema."""
        self._stats = self._meta.schema_df[['TableName', 'ColumnName', 'Cardinality']].copy()
        self._stats = self._stats.assign(
            Dictionary=self._meta.schema_df['Dictionary'].map(self._get_file_size_from_log),
            HashIndex=self._meta.schema_df['HIDX'].map(self._get_file_size_from_log),
            DataSize=self._meta.schema_df['IDF'].map(self._get_file_size_from_log),
            ModifiedTime=self._meta.schema_df['ModifiedTime'].apply(
                lambda x: WINDOWS_EPOCH_START + datetime.timedelta(seconds=x / 1e7)),
            StructureModifiedTime=self._meta.schema_df['StructureModifiedTime'].apply(
                lambda x: WINDOWS_EPOCH_START + datetime.timedelta(seconds=x / 1e7))
        )

    def _get_file_size_from_log(self, file_name):
        """Utility to get the size of a file from the log using 'FileName'."""
        file_ref = next((x for x in self._data_model.file_log if x['FileName'] == file_name), None)
        return file_ref['Size'] if file_ref else 0 
    
    @property
    def metadata(self):
        return self._meta
    
    @property
    def stats(self):
        return self._stats
    
    @property
    def size(self):
        return sum([x['Size'] for x in self._data_model.file_log])
    
    @property
    def schema(self):
        return  pd.DataFrame({
            'TableName': self._meta.schema_df['TableName'],
            'ColumnName': self._meta.schema_df['ColumnName'],
            'PandasDataType': self._meta.schema_df['DataType'].map(AMO_PANDAS_TYPE_MAPPING).fillna('object'),
        })

    @property   
    def tables(self):
        return self._meta.schema_df['TableName'].unique()


# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO


if getattr(kaitaistruct, 'API_VERSION', (0, 9)) < (0, 9):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class ColumnDataIdf(KaitaiStruct):
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._read()

    def _read(self):
        self.segments = []
        i = 0
        while not self._io.is_eof():
            self.segments.append(ColumnDataIdf.Segment(self._io, self, self._root))
            i += 1


    class Segment(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.primary_segment_size = self._io.read_u8le()
            self.primary_segment = []
            for i in range(self.primary_segment_size):
                self.primary_segment.append(ColumnDataIdf.SegmentEntry(self._io, self, self._root))

            self.sub_segment_size = self._io.read_u8le()
            self.sub_segment = []
            for i in range(self.sub_segment_size):
                self.sub_segment.append(self._io.read_u8le())



    class SegmentEntry(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.data_value = self._io.read_u4le()
            self.repeat_value = self._io.read_u4le()




import xml.etree.ElementTree as ET

class BackupLogHeader:
    def __init__(self, xml_string):
        # header is always one page (4096 bytes) in size and padded with zeros between the last header element and the end of the page
        trimmed_xml_string = xml_string.decode('utf-16').rstrip('\x00')
        root = ET.fromstring(trimmed_xml_string)
        self.BackupRestoreSyncVersion = int(root.findtext("BackupRestoreSyncVersion"))
        self.Fault = root.findtext("Fault") == "true"
        self.faultcode = int(root.findtext("faultcode"))
        self.ErrorCode = root.findtext("ErrorCode") == "true"
        self.EncryptionFlag = root.findtext("EncryptionFlag") == "true"
        self.EncryptionKey = int(root.findtext("EncryptionKey"))
        self.ApplyCompression = root.findtext("ApplyCompression") == "true"
        self.m_cbOffsetHeader = int(root.findtext("m_cbOffsetHeader"))
        self.DataSize = int(root.findtext("DataSize"))
        self.Files = int(root.findtext("Files"))
        self.ObjectID = root.findtext("ObjectID")
        self.m_cbOffsetData = int(root.findtext("m_cbOffsetData"))

    @classmethod
    def from_xml_string(cls, xml_string):
        trimmed_xml_string = xml_string.decode('utf-16').rstrip('\x00')
        return cls(trimmed_xml_string)



import ctypes
import platform
import zipfile
import os
from .abf import parser
from .abf.data_model import DataModel


class PbixUnpacker:
    def __init__(self, file_path):
        self.file_path = file_path

        # Attributes populated during unpacking
        self._data_model = DataModel(file_log=[], decompressed_data=b'')
        
        # Setup library
        self.__setup_library()
        
        # Trigger unpacking upon instantiation
        self.__unpack()

    def __setup_library(self):
        # Get the directory of the current file
        current_dir = os.path.dirname(os.path.abspath(__file__))

        # Determine the path to the shared library based on the platform
        if platform.system() == "Windows":
            self.lib_path = os.path.join(current_dir, 'lib', 'libxpress9.dll')
        elif platform.system() == "Linux":
            self.lib_path = os.path.join(current_dir, 'lib', 'libxpress9.so')
        elif platform.system() == "Darwin":
            self.lib_path = os.path.join(current_dir, 'lib', 'libxpress9.dylib')
        else:
            raise RuntimeError("Unsupported platform")

        # Load the shared library
        self.lib = ctypes.CDLL(self.lib_path)

        # Define the function signatures
        self.lib.Initialize.argtypes = []
        self.lib.Initialize.restype = ctypes.c_bool

        self.lib.Decompress.argtypes = [ctypes.POINTER(ctypes.c_ubyte), ctypes.c_int, ctypes.POINTER(ctypes.c_ubyte), ctypes.c_int]
        self.lib.Decompress.restype = ctypes.c_uint

        self.lib.Terminate.argtypes = []
        self.lib.Terminate.restype = None

    def __unpack(self):
        # Initialize the library
        result = self.lib.Initialize()
        if result:
            raise RuntimeError("Failed to initialize the library")
        
        with zipfile.ZipFile(self.file_path, 'r') as zip_ref:
            # Open the DataModel file within the ZIP
            with zip_ref.open('DataModel') as data_model_in_pbix:

                all_decompressed_data = bytearray()
                total_size = data_model_in_pbix.seek(0, 2)  # Get total size of file
                data_model_in_pbix.seek(102)  # Signature: This backup was created using Xpress9 compression.

                while data_model_in_pbix.tell() < total_size:
                    uncompressed_size = int.from_bytes(data_model_in_pbix.read(4), 'little')  # Read uint32 for uncompressed size
                    compressed_size = int.from_bytes(data_model_in_pbix.read(4), 'little')  # Read uint32 for compressed size
                    compressed_data = data_model_in_pbix.read(compressed_size)

                    # Create ctypes buffers
                    compressed_buffer = (ctypes.c_ubyte * compressed_size)(*compressed_data)
                    decompressed_buffer = (ctypes.c_ubyte * uncompressed_size)()

                    # Decompress the data
                    decompressed_size = self.lib.Decompress(compressed_buffer, compressed_size, decompressed_buffer, uncompressed_size)
                    if decompressed_size != uncompressed_size:
                        raise RuntimeError(f"Expected {uncompressed_size} bytes after decompression, but got {decompressed_size} bytes")

                    # Append decompressed data to all_decompressed_data
                    all_decompressed_data.extend(decompressed_buffer)

        # Terminate the library
        self.lib.Terminate()

        # Populate the byte array of the data bundle
        self._data_model.decompressed_data = all_decompressed_data

        # Parse the decompressed data
        abf_parser = parser.AbfParser(self._data_model)


    @property
    def data_model(self):
        return self._data_model

    @data_model.setter
    def data_model(self, value):
        if isinstance(value, DataModel):
            self._data_model = value
        else:
            raise ValueError("Expected an instance of DataModel.")


from .abf.data_model import DataModel
import datetime

# ---------- CONSTANTS ----------
AMO_PANDAS_TYPE_MAPPING = {
    2: 'string',
    6: 'Int64',
    8: 'Float64',
    9: 'datetime64[ns]',
    10: 'decimal.Decimal',
    11: 'bool',
    17: 'bytes'
}

# Windows epoch start date
WINDOWS_EPOCH_START = datetime.datetime(1601, 1, 1)

# ---------- UTILITY FUNCTIONS ----------
def get_data_slice(data_model:DataModel, file_name:str) -> bytes:
    """Gets a data slice based on a file name from the file log."""
    file_ref = next((x for x in data_model.file_log if x['FileName'] == file_name), None)
    if not file_ref:
        raise ValueError(f"File reference not found for filename: {file_name}.")
    # if error_code trim last 4 bytes
    if data_model.error_code:
        return data_model.decompressed_data[file_ref['m_cbOffsetHeader']:file_ref['m_cbOffsetHeader'] + file_ref['Size']-4]
    else:
        return data_model.decompressed_data[file_ref['m_cbOffsetHeader']:file_ref['m_cbOffsetHeader'] + file_ref['Size']]



# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO


if getattr(kaitaistruct, 'API_VERSION', (0, 9)) < (0, 9):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class IdfmetaParser(KaitaiStruct):
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._read()

    def _read(self):
        self.blocks = IdfmetaParser.Block(self._io, self, self._root)

    class CPElement(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.cp_tag = self._io.read_bytes(6)
            if not self.cp_tag == b"\x3C\x31\x3A\x43\x50\x00":
                raise kaitaistruct.ValidationNotEqualError(b"\x3C\x31\x3A\x43\x50\x00", self.cp_tag, self._io, u"/types/c_p_element/seq/0")
            self.version_one = self._io.read_u8le()
            self.cs = IdfmetaParser.CS0Element(self._io, self, self._root)
            self.cp_tag_end = self._io.read_bytes(6)
            if not self.cp_tag_end == b"\x43\x50\x3A\x31\x3E\x00":
                raise kaitaistruct.ValidationNotEqualError(b"\x43\x50\x3A\x31\x3E\x00", self.cp_tag_end, self._io, u"/types/c_p_element/seq/3")


    class CS1Element(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.cs_tag = self._io.read_bytes(6)
            if not self.cs_tag == b"\x3C\x31\x3A\x43\x53\x00":
                raise kaitaistruct.ValidationNotEqualError(b"\x3C\x31\x3A\x43\x53\x00", self.cs_tag, self._io, u"/types/c_s1_element/seq/0")
            self.count_bit_packed = self._io.read_u8le()
            self.blob_with9_zeros = self._io.read_bytes(9)
            self.cs_tag_end = self._io.read_bytes(6)
            if not self.cs_tag_end == b"\x43\x53\x3A\x31\x3E\x00":
                raise kaitaistruct.ValidationNotEqualError(b"\x43\x53\x3A\x31\x3E\x00", self.cs_tag_end, self._io, u"/types/c_s1_element/seq/3")


    class CS0Element(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.cs_tag = self._io.read_bytes(6)
            if not self.cs_tag == b"\x3C\x31\x3A\x43\x53\x00":
                raise kaitaistruct.ValidationNotEqualError(b"\x3C\x31\x3A\x43\x53\x00", self.cs_tag, self._io, u"/types/c_s0_element/seq/0")
            self.records = self._io.read_u8le()
            self.one = self._io.read_u8le()
            self.a_b_a_5_a = self._io.read_u4le()
            self.iterator = self._io.read_u4le()
            self.bookmark_bits_1_2_8 = self._io.read_u8le()
            self.storage_alloc_size = self._io.read_u8le()
            self.storage_used_size = self._io.read_u8le()
            self.segment_needs_resizing = self._io.read_u1()
            self.compression_info = self._io.read_u4le()
            self.ss = IdfmetaParser.SSElement(self._io, self, self._root)
            self.has_bit_packed_sub_seg = self._io.read_u1()
            self.cs = IdfmetaParser.CS1Element(self._io, self, self._root)
            self.cs_tag_end = self._io.read_bytes(6)
            if not self.cs_tag_end == b"\x43\x53\x3A\x31\x3E\x00":
                raise kaitaistruct.ValidationNotEqualError(b"\x43\x53\x3A\x31\x3E\x00", self.cs_tag_end, self._io, u"/types/c_s0_element/seq/13")


    class SDOsElement(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.sdos_tag = self._io.read_bytes(8)
            if not self.sdos_tag == b"\x3C\x31\x3A\x53\x44\x4F\x73\x00":
                raise kaitaistruct.ValidationNotEqualError(b"\x3C\x31\x3A\x53\x44\x4F\x73\x00", self.sdos_tag, self._io, u"/types/s_d_os_element/seq/0")
            self.csdos = IdfmetaParser.CSDOsElement(self._io, self, self._root)
            self.sdos_tag_end = self._io.read_bytes(8)
            if not self.sdos_tag_end == b"\x53\x44\x4F\x73\x3A\x31\x3E\x00":
                raise kaitaistruct.ValidationNotEqualError(b"\x53\x44\x4F\x73\x3A\x31\x3E\x00", self.sdos_tag_end, self._io, u"/types/s_d_os_element/seq/2")


    class SSElement(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.ss_tag = self._io.read_bytes(6)
            if not self.ss_tag == b"\x3C\x31\x3A\x53\x53\x00":
                raise kaitaistruct.ValidationNotEqualError(b"\x3C\x31\x3A\x53\x53\x00", self.ss_tag, self._io, u"/types/s_s_element/seq/0")
            self.distinct_states = self._io.read_u8le()
            self.min_data_id = self._io.read_u4le()
            self.max_data_id = self._io.read_u4le()
            self.original_min_segment_data_id = self._io.read_u4le()
            self.r_l_e_sort_order = self._io.read_s8le()
            self.row_count = self._io.read_u8le()
            self.has_nulls = self._io.read_u1()
            self.r_l_e_runs = self._io.read_u8le()
            self.others_r_l_e_runs = self._io.read_u8le()
            self.ss_tag_end = self._io.read_bytes(6)
            if not self.ss_tag_end == b"\x53\x53\x3A\x31\x3E\x00":
                raise kaitaistruct.ValidationNotEqualError(b"\x53\x53\x3A\x31\x3E\x00", self.ss_tag_end, self._io, u"/types/s_s_element/seq/10")


    class Block(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.cp = IdfmetaParser.CPElement(self._io, self, self._root)
            if self._io.size() > self._io.pos():
                self.sdos = IdfmetaParser.SDOsElement(self._io, self, self._root)


    class CSDOs1Element(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.csdos_tag = self._io.read_bytes(9)
            if not self.csdos_tag == b"\x3C\x31\x3A\x43\x53\x44\x4F\x73\x00":
                raise kaitaistruct.ValidationNotEqualError(b"\x3C\x31\x3A\x43\x53\x44\x4F\x73\x00", self.csdos_tag, self._io, u"/types/c_s_d_os1_element/seq/0")
            self.sub_segment_offset = self._io.read_u8le()
            self.sub_segment_size = self._io.read_u8le()
            self.csdos_tag_end = self._io.read_bytes(9)
            if not self.csdos_tag_end == b"\x43\x53\x44\x4F\x73\x3A\x31\x3E\x00":
                raise kaitaistruct.ValidationNotEqualError(b"\x43\x53\x44\x4F\x73\x3A\x31\x3E\x00", self.csdos_tag_end, self._io, u"/types/c_s_d_os1_element/seq/3")


    class CSDOsElement(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.csdos_tag = self._io.read_bytes(9)
            if not self.csdos_tag == b"\x3C\x31\x3A\x43\x53\x44\x4F\x73\x00":
                raise kaitaistruct.ValidationNotEqualError(b"\x3C\x31\x3A\x43\x53\x44\x4F\x73\x00", self.csdos_tag, self._io, u"/types/c_s_d_os_element/seq/0")
            self.zero_c_s_d_o = self._io.read_u8le()
            self.primary_segment_size = self._io.read_u8le()
            self.csdos = IdfmetaParser.CSDOs1Element(self._io, self, self._root)
            self.csdos_tag_end = self._io.read_bytes(9)
            if not self.csdos_tag_end == b"\x43\x53\x44\x4F\x73\x3A\x31\x3E\x00":
                raise kaitaistruct.ValidationNotEqualError(b"\x43\x53\x44\x4F\x73\x3A\x31\x3E\x00", self.csdos_tag_end, self._io, u"/types/c_s_d_os_element/seq/4")


    @property
    def bit_width(self):
        if hasattr(self, '_m_bit_width'):
            return self._m_bit_width

        self._m_bit_width = ((36 - self.blocks.cp.cs.a_b_a_5_a) + self.blocks.cp.cs.iterator)
        return getattr(self, '_m_bit_width', None)





from setuptools import setup, find_packages

setup(
    name='pbixray',
    version='0.1.21',
    packages=find_packages(),
    install_requires=[
        'kaitaistruct',
        'pandas',
        'apsw'
    ],
    include_package_data=True,
    package_data={
        'pbixray': ['lib/*.dll', 'lib/*.so', 'lib/*.dylib'],
    },
    author="Igor Cotruta",
    description="A Python library to parse and analyze PBIX files used with Microsoft Power BI.",
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url="https://github.com/Hugoberry/pbixray",
    license="MIT",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Information Technology",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Visualization",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



import xml.etree.ElementTree as ET

class BackupLog:
    def __init__(self, xml_string, error_code):
        #if error_code trim last 4 bytes
        if error_code:
            xml_string = xml_string[:-4]
        trimmed_xml_string = xml_string.decode('utf-16')
        root = ET.fromstring(trimmed_xml_string)
        self.BackupRestoreSyncVersion = root.findtext("BackupRestoreSyncVersion")
        self.ServerRoot = root.findtext("ServerRoot")
        self.SvrEncryptPwdFlag = root.findtext("SvrEncryptPwdFlag") == "true"
        self.ServerEnableBinaryXML = root.findtext("ServerEnableBinaryXML") == "true"
        self.ServerEnableCompression = root.findtext("ServerEnableCompression") == "true"
        self.CompressionFlag = root.findtext("CompressionFlag") == "true"
        self.EncryptionFlag = root.findtext("EncryptionFlag") == "true"
        self.ObjectName = root.findtext("ObjectName")
        self.ObjectId = root.findtext("ObjectId")
        self.Write = root.findtext("Write")
        self.OlapInfo = root.findtext("OlapInfo") == "true"
        self.Collations = [collation.text for collation in root.findall("Collations/Collation")]
        self.Languages = [int(lang.text) for lang in root.findall("Languages/Language")]
        self.FileGroups = [FileGroup.from_xml(filegroup) for filegroup in root.findall("FileGroups/FileGroup")]

class FileGroup:
    @classmethod
    def from_xml(cls, element):
        fileGroup = cls()
        fileGroup.Class = int(element.findtext("Class"))
        fileGroup.ID = element.findtext("ID")
        fileGroup.Name = element.findtext("Name")
        fileGroup.ObjectVersion = int(element.findtext("ObjectVersion"))
        fileGroup.PersistLocation = int(element.findtext("PersistLocation"))
        fileGroup.PersistLocationPath = element.findtext("PersistLocationPath")
        fileGroup.StorageLocationPath = element.findtext("StorageLocationPath")
        fileGroup.ObjectID = element.findtext("ObjectID")
        fileGroup.FileList = [BackupFile.from_xml(backupfile) for backupfile in element.findall("FileList/BackupFile")]
        return fileGroup

class BackupFile:
    @classmethod
    def from_xml(cls, element):
        backupFile = cls()
        backupFile.Path = element.findtext("Path")
        backupFile.StoragePath = element.findtext("StoragePath")
        backupFile.LastWriteTime = int(element.findtext("LastWriteTime"))
        backupFile.Size = int(element.findtext("Size"))
        return backupFile
